'''Dymaxion Projection Conversion Subroutines'''
import math
from functools import lru_cache
import numba
import numpy as np
import time

from . import constants

@numba.njit(fastmath=True)
def magnitude(vec):
    '''N-Dimentional Vector Magnitude'''
    acc = 0
    for val in vec:
        acc += val**2
    return math.sqrt(acc)

@numba.njit(fastmath=True)
def euclidean(vec_a, vec_b):
    '''N-Dimensional Vector L2 Norm'''
    acc = 0
    for idx in range(len(vec_a)):
        acc += (vec_a[idx] - vec_b[idx])**2
    return math.sqrt(acc)

### Dymax Conversion Main Routine
@lru_cache(maxsize=2**12)
def lonlat2dymax(lon, lat, getlcd=False):
    '''
    Lon Lat 2 Dymax XY

    This is the primary reason this whole package exists.

    Parameters
    ----------
    lon : float
        Longitude in radians.
    lat : float
        Latitude in radians.
    getlcd : bool, optional(default False)
        Set to return the LCD triangle index. You get it for free when computing position in triangle.

    Returns
    -------
    x_pos : float
        X position in dymaxion coordinates.
    y_pos : float
        Y position in dymaxion coordinates.
    lcd : int, optional
        Lowest-Common-Denominator sub-triangle index.

    Example
    -------
    >>> vert = lonlat2dymax(-77.0367, 38.8951)
    >>> 'x={:.8f}, y={:.8f}'.format(*vert)
    'x=3.30326834, y=1.53381487'
    '''
    # Convert the given(lon, lat) coordinate into spherical
    # polar coordinates(r, theta, phi) with radius=1.
    # Angles are given in radians, NOT degrees.
    theta, phi = lonlat2spherical(lon, lat)

    # convert the spherical polar coordinates into cartesian
    # (x, y, z) coordinates.
    XYZ = spherical2cartesian(theta, phi)
    # determine which of the 20 spherical icosahedron triangles
    # the given point is in and the LCD triangle.
    tri, lcd = fuller_triangle(XYZ)

    # Determine the corresponding Fuller map plane(x, y) point
    x_pos, y_pos = dymax_point(tri, lcd, XYZ)

    if getlcd: return x_pos, y_pos, lcd
    else:      return x_pos, y_pos

### Dymax Conversion Subroutines
def vert2dymax(vert, vertset, push=.9999):
    '''
    Convert Vertex Index to XY Position

    We need to 'nudge' the point a little bit into the triangle.
    Without the nudge, the vertices would be exactly between dymaxion faces and wouldn't make sense for plotting.
    Hence we do a weighted average with point idx having a massive weight.

    Example
    -------
    >>> vert = vert2dymax(3, constants.vert_indices[1])
    >>> 'x={:.8f}, y={:.8f}'.format(*vert)
    'x=2.00000033, y=0.86617338'
    '''
    XYZ = np.zeros(3)
    for idx in range(3):
        if vertset[idx] == vert:
            XYZ += constants.vertices[vert] * push
        else:
            XYZ += constants.vertices[vertset[idx]] * (1-push)

    ### Determine the corresponding Fuller map plane(x, y) point
    tri, hlcd = fuller_triangle(XYZ)
    x_pos, y_pos = dymax_point(tri, hlcd, XYZ)
    return x_pos, y_pos

def face2dymax(face_idx, push=.9999, atomic=False):
    '''
    Convert Icosahedron Face to (4) XY Vertices

    Parameters
    ----------
    face_idx : int
        Dymaxion face index.
    push : float
        Multiplier distance from vertex to center. Has the effect of compressing
        the face toward the center for easy plotting usage.
    atomic will draw the LCD subtriangles

    Returns
    -------
    points : ndarray of float
        Vertices correspointing to the face index requested. Normally this is
        4 vertics, but if atomic will return 7 vertices.

    Example
    -------
    >>> verts = face2dymax(1, push=.75)
    >>> for vdx, vert in enumerate(verts):
    ...     print('v{} x={:.8f}, y={:.8f}'.format(vdx, vert[0], vert[1]))
    v0 x=2.35304556, y=1.64720662
    v1 x=1.64695413, y=1.64720662
    v2 x=2.00000025, y=1.03571383
    v3 x=2.35304556, y=1.64720662
    '''
    if atomic:
        points = np.zeros((6+1, 2))
        for jdx in range(6):
            if not jdx % 2: XYZ = constants.vertices[constants.vert_indices[face_idx, jdx//2]] # Normal Vertex
            else:
                up = constants.vertices[constants.vert_indices[face_idx, (jdx//2+1)%3]]
                down = constants.vertices[constants.vert_indices[face_idx, (jdx//2+2)%3]]
                XYZ = np.mean([up, down], axis=0)
            XYZ = XYZ * push + constants.XYZcenters[face_idx] * (1-push)
            tri, hlcd = fuller_triangle(XYZ)
            points[jdx] = dymax_point(tri, hlcd, XYZ)
    else:
        points = np.zeros((3+1, 2))
        for jdx in range(3):
            XYZ = constants.vertices[constants.vert_indices[face_idx, jdx]] * push + constants.XYZcenters[face_idx] * (1-push)
            tri, hlcd = fuller_triangle(XYZ)
            points[jdx] = dymax_point(tri, hlcd, XYZ)

    points[-1] = points[0] # Loop Back to Start
    return points

def lonlat2spherical(lon, lat):
    '''
    Convert (lon, lat) point into spherical polar coordinates with radius=1.
    Angles are given in radians.
    note: Not on WGS84 Ellipsoid

    >>> theta, phi = lonlat2spherical(179, 89)
    >>> 'theta={:.6f}, phi={:.6f}'.format(theta, phi)
    'theta=0.017453, phi=3.124139'
    '''
    h_theta = 90 - lat
    h_phi = lon
    if lon < 0: h_phi = lon + 360
    theta = math.radians(h_theta)
    phi = math.radians(h_phi)
    return theta, phi

def spherical2cartesian(theta, phi):
    '''
    Covert spherical polar coordinates to cartesian coordinates.
    Input angles in radians, output as unit vector.

    Note that numba doesn't speed this one up.

    >>> spherical2cartesian(math.pi/2, math.pi)
    (-1.0, 1.2246467991473532e-16, 6.123233995736766e-17)
    '''
    x = math.sin(theta) * math.cos(phi)
    y = math.sin(theta) * math.sin(phi)
    z = math.cos(theta)
    return x, y, z

def cartesian2spherical(XYZ):
    '''
    Convert Cartesian to Spherical (Non-WGS84)
    Takes a [X,Y,Z] unit vector as input.
    (theta, phi) ~ (lon, lat)

    >>> cartesian2spherical([0.131, -0.84, 0.525])
    (-1.4160901241763815, 1.0180812136981134)
    '''
    phi = math.acos(XYZ[2])
    theta = math.atan2(XYZ[1], XYZ[0])
    return theta, phi

def fuller_triangle(XYZ):
    '''
    Determine which major icosahedron triangle
    and minor lowest common dinominator triangle
    the XYZ point is in. (6 LCDs per Triangle)

    Parameters
    ----------
    xyz : tuple of floats
        Cartesian coordinate position.

    Returns
    -------
    h_tri : int
        Dymaxion triangle index.
    h_lcd : int
        Dymaxion lowest-common-denominator triangle index.

    Example
    -------
    >>> fuller_triangle([-1, 0, 0])
    (10, 2)
    '''
    h_tri = -1
    h_dist1 = np.inf

    # Which triangle face center is the closest to the given point
    # is the triangle in which the given point is in.
    for idx in range(constants.facecount):
        h = constants.XYZcenters[idx] - XYZ
        h_dist2 = magnitude(h)
        if h_dist2 < h_dist1:
            h_tri = idx
            h_dist1 = h_dist2

    # Now the LCD triangle is determined.
    v1, v2, v3 = constants.vert_indices[h_tri]
    h_dist1 = euclidean(XYZ, constants.vertices[v1])
    h_dist2 = euclidean(XYZ, constants.vertices[v2])
    h_dist3 = euclidean(XYZ, constants.vertices[v3])

    if   h_dist1 <= h_dist2 <= h_dist3: h_lcd = 0
    elif h_dist1 <= h_dist3 <= h_dist2: h_lcd = 5
    elif h_dist2 <= h_dist1 <= h_dist3: h_lcd = 1
    elif h_dist2 <= h_dist3 <= h_dist1: h_lcd = 2
    elif h_dist3 <= h_dist1 <= h_dist2: h_lcd = 4
    elif h_dist3 <= h_dist2 <= h_dist1: h_lcd = 3
    return h_tri, h_lcd

def dymax_point(tri, lcd, XYZ):
    '''
    In order to rotate the given point into the template spherical
    triangle, we need the spherical polar coordinates of the center
    of the face and one of the face vertices. So set up which vertex
    to use.

    Parameters
    ----------
    tri : int
        Dymaxion face index where we want to be.
    lcd : int
        Dymaxion sub-triangle where we want to be.
    XYZ : tuple of floats
        Pseudo-ECEF coordinate that will be projected to dymaxion.

    Returns
    -------
    pointx : float
        X position for dymaxion projection.
    pointy : float
        Y position for dymaxion projection.

    Example
    -------
    >>> vert = dymax_point(10, 2, [-1.0, 0, 0])
    >>> 'x={:.8f}, y={:.8f}'.format(*vert)
    'x=3.50247081, y=0.09535516'
    '''
    v1 = constants.vert_indices[tri][0]

    h0XYZ = XYZ
    h1XYZ = constants.vertices[v1]

    theta, phi = cartesian2spherical(constants.XYZcenters[tri])

    axis = 2
    h0XYZ = rotate3d(axis, theta, h0XYZ)
    h1XYZ = rotate3d(axis, theta, h1XYZ)

    axis = 1
    h0XYZ = rotate3d(axis, phi, h0XYZ)
    h1XYZ = rotate3d(axis, phi, h1XYZ)

    theta, phi = cartesian2spherical(h1XYZ)
    theta = theta - np.pi/2

    axis = 2
    h0XYZ = rotate3d(axis, theta, h0XYZ)

    ### exact transformation equations
    gz = math.sqrt(1 - h0XYZ[0]**2 - h0XYZ[1]**2)
    gs = math.sqrt(5 + 2 * math.sqrt(5)) / (gz * math.sqrt(15))

    gxp = h0XYZ[0] * gs
    gyp = h0XYZ[1] * gs

    ga0p = 2 * gyp / math.sqrt(3) + (constants.gel / 3)
    ga1p = gxp - (gyp / math.sqrt(3)) +  (constants.gel / 3)
    ga2p = (constants.gel / 3) - gxp - (gyp / math.sqrt(3))

    ga0 = constants.gt + math.atan2(ga0p - 0.5 * constants.gel, constants.gdve)
    ga1 = constants.gt + math.atan2(ga1p - 0.5 * constants.gel, constants.gdve)
    ga2 = constants.gt + math.atan2(ga2p - 0.5 * constants.gel, constants.gdve)

    gx = 0.5 * (ga1 - ga2)
    gy = (2 * ga0 - ga1 - ga2) / (2 * math.sqrt(3))

    ### Re-scale so plane triangle edge length is 1
    pointx = gx / constants.garc
    pointy = gy / constants.garc

    ### Move and Rotate as Appropriate
    # You can disable the special translations for uniform triangles
    if   tri == 8  and lcd < 4:
        xtranslate, ytranslate, rotation = constants.dymax_translate08_special
    elif tri == 15 and lcd < 3:
        xtranslate, ytranslate, rotation = constants.dymax_translate15_special
    else:
        xtranslate, ytranslate, rotation = constants.dymax_translate[tri]

    pointx, pointy = rotate2d(rotation, pointx, pointy)
    pointx += xtranslate
    pointy += ytranslate
    return pointx, pointy

def rotate2d(angle, pointx, pointy):
    '''
    Rotate a point orientation in XY-plane around Z
    This function obeys the right hand rule.

    >>> rotate2d(90,.5,1)
    (-1.0, 0.5000000000000001)
    '''
    ha = math.radians(angle)
    hx = pointx
    hy = pointy
    pointx = hx * math.cos(ha) - hy * math.sin(ha)
    pointy = hx * math.sin(ha) + hy * math.cos(ha)

    return pointx, pointy

def rotate3d(axis, alpha, XYZ, reverse=True):
    '''
    Rotate a 3-D point about the specified axis by alpha radians
    For some horrible reason, we are doing left hand rotation.
    reverse == left hand rotation, set to False for normal

    >>> rotate3d(0, np.pi/4, [.3, .5, .4])
    (0.3, 0.6363961030678928, -0.07071067811865467)
    '''
    if reverse: alpha = -alpha

    if axis == 0:
        # Rotate around X
        XYZ = (XYZ[0],
               XYZ[1] * math.cos(alpha) - XYZ[2] * math.sin(alpha),
               XYZ[1] * math.sin(alpha) + XYZ[2] * math.cos(alpha))

    elif axis == 1:
        # Rotate around Y
        XYZ = (XYZ[0] * math.cos(alpha) + XYZ[2] * math.sin(alpha),
               XYZ[1],
               -XYZ[0] * math.sin(alpha) + XYZ[2] * math.cos(alpha))

    elif axis == 2:
        # Rotate around Z
        XYZ = (XYZ[0] * math.cos(alpha) - XYZ[1] * math.sin(alpha),
               XYZ[0] * math.sin(alpha) + XYZ[1] * math.cos(alpha),
               XYZ[2])

    return XYZ

def raytrace(x_loc, y_loc, poly):
    '''
    Determine if position (x_loc, y_loc) is inside polygon.

    Examples
    --------
    >>> raytrace(2.01, .5, [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]])
    False
    >>> raytrace(0.99, .5, [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]])
    True
    '''
    psize = len(poly)
    inside = False
    p2x = 0.
    p2y = 0.
    xints = 0.
    p1x, p1y = poly[0]
    for pdx in range(psize + 1):
        p2x, p2y = poly[pdx % psize]
        if y_loc > min(p1y, p2y):
            if y_loc <= max(p1y, p2y):
                if x_loc <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y_loc-p1y) * (p2x-p1x) / (p2y-p1y) + p1x
                    if p1x == p2x or x_loc <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y
    return inside

def benchmark(verbose=True):
    '''
    simple unique point benchmark

    on i7-8550U, points/sec
    v1.0.0: 13600
    v1.1.0: 28000
    v1.1.1: 18300
    v1.1.3: 16000
    '''
    lon_res = 1000
    lat_res = 100
    lons = np.linspace(-180, 180, lon_res)
    lats = np.linspace(-90, 90, lat_res)
    start = time.time()
    for lat in lats:
        for lon in lons:
            _ = lonlat2dymax(lon, lat)
    if verbose:
        print(':: mapped {:d} unique points to dymax projection @ {:.1f} pts/sec [{:.1f} secs total]'.format(
            lon_res * lat_res,
            (lon_res * lat_res) / (time.time()-start),
            time.time()-start))

### Determine (X,Y) Projection Coordinates for Dymaxion Triangle Centers
dymax_centers = np.zeros((constants.facecount, 2))
for fdx in range(constants.facecount):
    tri, hlcd = fuller_triangle(constants.XYZcenters[fdx])
    dymax_centers[fdx] = dymax_point(tri, hlcd, constants.XYZcenters[fdx])

if __name__ == '__main__':
    benchmark()



'''Routines to load shoreline data'''
import lzma
import numpy as np
import os
import pkg_resources
import struct
import time

from . import convert

PKG_DATA = pkg_resources.resource_filename('dymax', 'data') + os.path.sep

def load_gshhs_xz(filename):
    '''
    Convert GSHHS Binary File to Numpy Arrays

    Written for GSHHS v2.3.7 utilizing API v2.0
    For some insane reason in this version they didn't wrap the coastlines
    around the longitude discontinutiy correctly.

    Files were created by using:
    for res in ['c', 'l', 'i', 'h', 'f']:
        with open('gshhs_{}.b'.format(res), 'rb') as raw:
            with open('gshhs_{}.xz'.format(res), 'wb') as out:
                out.write(lzma.compress(raw.read()))

    Returns
    -------
    headers : ndarray of int32
        n : int
            number of points in this coastline
        flag : int
            (5 items) = level + version << 8 + greenwich << 16 + source << 24 + river << 25
            low byte:    level = flag & 255: Values: 1 land, 2 lake, 3 island_in_lake, 4 pond_in_island_in_lake
            2nd byte:    version = (flag >> 8) & 255: Values: Should be 12 for GSHHG release 12 (i.e., version 2.2)
            3rd byte:    greenwich = (flag >> 16) & 1: Values: Greenwich is 1 if Greenwich is crossed
            4th byte:    source = (flag >> 24) & 1: Values: 0 = CIA WDBII, 1 = WVS
            4th byte:    river = (flag >> 25) & 1: Values: 0 = not set, 1 = river-lake and level = 2
        area_full : int
            Area of original full-resolution polygon in 1/10 km^2.
        container: int
            Index of container polygon that encloses this polygon (-1 if none)
        ancestor : int
            Index of ancestor polygon in the full resolution set that was the source of this polygon (-1 if none)
    wvs : list of ndarray of int32
        lon : float
            WGS84 Longitude in degrees from (-180, 180)
        lat : float
            WGS84 Latitude in degrees from (-90, 90)
    '''
    with lzma.open(filename, 'rb') as handle:
        # world vector shoreline
        # lon, lat
        coasts = []
        # headers
        # id, n, flag, west, east, south, north, area, area_full, container, ancestor
        headers = []
        while True:
            try:
                header = struct.unpack(
                    '>11i',
                    handle.read(11*4))
                numbytes = header[1]
                coast = struct.unpack(
                    '>{}i'.format(numbytes*2),
                    handle.read(numbytes*2*4))
                coast = np.array(coast, dtype=np.int32).reshape((numbytes, 2))
                # dump
                header = np.array(header[1:3] + header[8:], dtype=np.int32)
                # convert microdegrees to degrees
                coast = coast.astype(np.float32) / 1e6
                # chang lon from (-180, 360) to (-180, 180)
                lon = np.radians(coast[:, 0])
                lon = np.degrees(np.arctan2(np.sin(lon), np.cos(lon)))
                coast[:, 0] = lon
                coasts += [coast]
                headers += [header]
            except struct.error:
                print('{} EOF'.format(filename))
                break
    headers = np.vstack(headers)
    return headers, coasts

def get_coastlines(resolution='c'):
    '''
    Return Dymax Lands

    Parameters
    ----------
    resolution : string
        Resolutions are valid in the following set:
            Crude Resolution (25 km) 'c'
            Low Resolution (5 km) 'l'
            Intermediate Resolution (1 km) 'i'
            High Resolution (0.2 km) 'h'
            Full Resolution (0.04 km) 'f'
    filter : list
        Filter for flag. See load_gshhs_xz for details.

    Returns
    -------
    coasts : list of 2-D ndarrays
        Each island in list contains an array of N points. Each point is a
        (lon, lat) pair of WGS84 coordinates.
    dymax_coasts : list of list of floats
        Each island in list contains an array of N points. Each point is a
        (x_pos, y_pos) pair of dymax coordinates.
    '''
    dymax_coasts = []
    headers, coasts = load_gshhs_xz(os.path.join(PKG_DATA,'gshhs_{}.xz'.format(resolution)))
    for hdx, header in enumerate(headers):
        dymax_coast = []
        # only consider land, not islands or lakesd
        for cdx in range(header[0]):
            lon, lat = coasts[hdx][cdx]
            dymax_coast += [convert.lonlat2dymax(lon, lat)]
        dymax_coasts += [dymax_coast]
    return coasts, dymax_coasts



#!/usr/bin/env python3
'''Dymaxion Projection Examples'''
import os
import time
from sys import stdout
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from PIL import Image, ImageOps

from . import convert
from . import constants
from . import io

def plot_triangles(save:bool=False, show:bool=True, dpi:int=300, verbose:bool=True):
    '''Draw Dymax Spherical Triangles'''
    plt.figure(figsize=(20, 12))
    for jdx in range(constants.facecount):
        if jdx in [8, 15]: continue
        points = convert.face2dymax(jdx, push=.95)
        xcenter, ycenter = convert.dymax_centers[jdx]
        plt.text(xcenter, ycenter, str(jdx), ha='center', va='center')
        plt.plot(points[:, 0], points[:, 1], lw=5, alpha=.5)
    plt.gca().set_xlim([0, 5.5])
    plt.gca().set_ylim([0, 2.6])
    plt.gca().set_aspect('equal')
    if verbose:
        print(':: plotted', constants.facecount, 'triangles')
    if save:
        plt.savefig('dymax_triangles.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def plot_triangles_meridians(resolution:str='c', save:bool=False, show:bool=True, dpi:int=300, verbose:bool=True):
    '''Draw Dymax Triangles, All countries, and Meridians'''
    lonlat_islands, dymax_islands = io.get_coastlines(resolution)
    n = 1000
    plt.figure(figsize=(20, 12))
    plt.title('Dymaxion Map Projection')

    ### Dymaxion Latitude Meridians
    lons = np.linspace(-180, 180, n)
    latgrid = np.linspace(-85, 85, 35)
    points = []
    start = time.time()
    for lat in latgrid:
        for lon in lons:
            points += [convert.lonlat2dymax(lon, lat)]
    if verbose:
        print(':: mapped {:d} points to dymax projection @ {:.1f} pts/sec [{:.1f} secs total]'.format(
            len(points), len(points)/(time.time()-start), time.time()-start))
    points = np.array(points)
    plt.plot(points[:, 0], points[:, 1], ',', c='k', alpha=.3)

    ### Dymaxion Longitude Meridians
    lats = np.linspace(-85, 85, n)
    longrid = np.linspace(-180, 175, 72)
    points = []
    start = time.time()
    for lon in longrid:
        for lat in lats:
            points += [convert.lonlat2dymax(lon, lat)]
    if verbose: print(':: mapped {:d} points to dymax projection @ {:.1f} pts/sec [{:.1f} secs total]'.format(
                    len(points), len(points)/(time.time()-start), time.time()-start))
    points = np.array(points)
    plt.plot(points[:, 0], points[:, 1], ',', c='k', alpha=.3)

    ### Dymaxion Face Tiles
    for jdx in range(constants.facecount):
        if jdx in [8, 15]: continue
        points = convert.face2dymax(jdx, push=.95)
        xcenter, ycenter = convert.dymax_centers[jdx]
        plt.text(
            xcenter, ycenter, str(jdx),
            size='x-large', ha='center', va='center')
        plt.plot(points[:, 0], points[:, 1], lw=5, alpha=.7)

    ### Draw Landmasses
    patches = []
    for island in dymax_islands:
        polygon = Polygon(np.array(island))#, closed=False, fill=False)
        patches.append(polygon)

    p = PatchCollection(patches, alpha=.3, linewidths=1, facecolors=None)
    colors = 100*np.random.random(len(patches))
    p.set_array(np.array(colors))
    plt.gca().add_collection(p)
    if verbose: print(':: plotted', len(patches), 'coastlines')
    plt.gca().set_xlim([0, 5.5])
    plt.gca().set_ylim([0, 2.6])
    plt.gca().set_aspect('equal')
    if save:
        plt.savefig('dymax_earthmeridianstriangles.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def plot_triangles_rectilinear(resolution:str='c', save:bool=False, show:bool=True, dpi:int=300, verbose:bool=True):
    lonlat_islands, dymax_islands = io.get_coastlines(resolution)
    plt.figure(figsize=(20, 12))
    plt.title('The dymax face polygons look super-fucked on a rectilinear projection')
    patches = []
    faces = []
    for island in lonlat_islands:
        polygon = Polygon(np.array(island), closed=False, fill=True)
        patches.append(polygon)

    for face in range(constants.facecount):
        derp = np.zeros((3, 2))
        for vtex in range(3):
            derp[vtex] = constants.lon_lat_verts[constants.vert_indices[face, vtex]]
        polygon = Polygon(derp, closed=False, fill=True)
        faces.append(polygon)

    colors = 100*np.random.random(len(patches))
    p = PatchCollection(patches, cmap=plt.cm.viridis, alpha=0.7, linewidths=0)
    f = PatchCollection(faces, cmap=plt.cm.viridis, alpha=0.3, linewidths=1)
    p.set_array(np.array(colors))
    f.set_array(np.array(colors))
    plt.gca().add_collection(p)
    plt.gca().add_collection(f)
    if verbose: print(':: plotted', len(patches), 'coastlines')
    plt.xlim(-180, 180)
    plt.ylim(-90, 90)
    if save:
        plt.savefig('dymax_rectilineartriangles.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def plot_lcd_triangles(verbose:bool=True, save:bool=False, show:bool=True, dpi:int=300, resolution:str='c'):
    '''Each Icosahedron Face has six sub-triangles that are splitting on.'''
    lonlat_islands, dymax_islands = io.get_coastlines(resolution)
    plt.figure(figsize=(20, 12))

    xs, ys = [], []
    lcds = []
    for _ in range(10000):
        lon = np.random.random()*360 - 180
        lat = np.random.random()*180 - 90
        x, y, lcd = convert.lonlat2dymax(lon, lat, getlcd=True)
        xs += [x]
        ys += [y]
        lcds += [lcd]

    ### Draw LCD Triangle Indices
    colorlist = 'rgcmyg'
    for val in range(10000):
        plt.text(
            xs[val], ys[val], str(lcds[val]),
            color=colorlist[lcds[val]], alpha=.5,
            ha='center', va='center', size='small')

    ### Draw Large Fuller Triangles
    for jdx in range(constants.facecount):
        if jdx in [8, 15]: continue # Edge Triangles
        points = convert.face2dymax(jdx, push=.95)
        xcenter, ycenter = convert.dymax_centers[jdx]
        plt.plot(points[:, 0], points[:, 1], lw=5, alpha=.5)

    ### Draw Fuller LCD Sub-Triangles
    for jdx in range(constants.facecount):
        if jdx in [8, 15]: continue # Edge Triangles
        points = convert.face2dymax(jdx, push=.999, atomic=True)
        xcenter, ycenter = convert.dymax_centers[jdx]
        plt.text(xcenter, ycenter, str(jdx), size='xx-large', ha='center', va='center')
        plt.plot(points[:, 0], points[:, 1], lw=1, alpha=1, color='k', ls='dotted')

    ### Draw Landmasses
    patches = []
    for island in dymax_islands:
        polygon = Polygon(np.array(island), closed=True, fill=True)
        patches.append(polygon)

    colors = 100*np.random.random(len(patches))
    p = PatchCollection(patches, cmap=plt.cm.viridis, alpha=.5, linewidths=0)
    p.set_array(np.array(colors))
    plt.gca().add_collection(p)
    if verbose: print(':: plotted', len(patches), 'coastlines')
    plt.gca().set_xlim([0, 5.5])
    plt.gca().set_ylim([0, 2.6])
    plt.gca().set_aspect('equal')

    if save:
        plt.savefig('dymax_earthsubtriangles.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def plot_grid(verbose:bool=True, save:bool=False, show:bool=True, dpi:int=300):
    '''Show Dymaxion Grid'''
    plt.figure(figsize=(20, 12))
    patches = []
    for zdx, vertset in enumerate(constants.vert_indices):
        if zdx in [8, 15]: continue # Edge Triangles
        x, y = [], []
        for _, vert in enumerate(vertset):
            xt, yt = convert.vert2dymax(vert, vertset)
            #print(xt,yt)
            x += [xt]
            y += [yt]
            #print(xt,yt,i,vert)
        #plt.plot(x,y,'k',lw=.1)
        patches.append(Polygon(np.array([x, y]).T, closed=False, fill=True))

    colors = 100*np.random.random(len(patches))
    p = PatchCollection(patches, cmap=plt.cm.viridis, alpha=1, linewidths=0.)
    p.set_array(np.array(colors))
    plt.gca().add_collection(p)
    if verbose: print(':: plotted', len(patches), 'coastlines')
    plt.gca().set_aspect('equal')
    plt.gca().set_xlim([0, 5.5])
    plt.gca().set_ylim([0, 2.6])
    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)
    plt.gca().axis('off')

    if save:
        plt.savefig('dymax_grid.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def plot_coastline_vectors(verbose=True, save=False, show=True, dpi=300, resolution='c'):
    '''Draw Landmasses Only, no Background'''
    lonlat_islands, dymax_islands = io.get_coastlines(resolution)

    patches = []
    for island in dymax_islands:
        polygon = Polygon(np.array(island), closed=True, fill=True)
        patches.append(polygon)

    plt.figure(figsize=(20, 12), frameon=False)
    colors = 100 * np.random.random(len(patches))
    p = PatchCollection(patches, cmap=plt.cm.viridis, alpha=1, linewidths=0.)
    p.set_array(np.array(colors))
    plt.gca().add_collection(p)
    if verbose: print(':: plotted', len(patches), 'coastlines')
    plt.gca().set_aspect('equal')
    plt.gca().set_xlim([0, 5.5])
    plt.gca().set_ylim([0, 2.6])
    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)
    plt.gca().axis('off')
    if save:
        plt.savefig('dymax_landmasses.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def convert_rectimage_2_dymaximage(inFilename, outFilename, verbose:bool=True, scale:float=300, speedup:int=1, save:bool=False, show:bool=True):
    '''
    Convert rectilinear image to dymax projection image.

    scale is number of pixels per dymax xy unit.
    How to calculate output scale:
        width = 160 # in cm
        resolution = 30 # in px/cm
        scale = (width * resolution) / 5.5
        __OR__
        final_size_in_pixels = (scale * 5.5, scale * 2.6)

    speedup gives a sparse preview of the output image and is specified as a time divisor.
    '''
    start = time.time()
    # Can be many different formats; 15 vertical and horizontal pixels per degree
    im = Image.open(inFilename)
    pix = im.load()
    if verbose: print(':: input image resolution =', im.size) # Get the width and hight of the image for iterating over

    ### LongLat2Dymax returns x = (0,5.5) and y=(0,2.6)
    dymax_xsize, dymax_ysize = int(5.5*scale), int(2.6*scale)
    dymaximg = Image.new('RGBA', (dymax_xsize, dymax_ysize), (255, 0, 0, 0)) # create a new transparent
    if verbose: print(':: output image resolution =', (dymax_xsize, dymax_ysize)) # Get the width and hight of the image for iterating over

    ### X and Y are indexed from topleft to bottom right
    if verbose: print(':: sweeping over Longitudes:')
    xsize, ysize = im.size
    for i, lon in enumerate(np.linspace(-180, 180, int(xsize/speedup), endpoint=True)):
        i *= speedup
        if i % 20 == 0:
            print('{:+07.2f} '.format(lon), end='')
            stdout.flush() # I would add flush=True to print, but thats only in python3.3+
        for j, lat in enumerate(np.linspace(90, -90, int(ysize/speedup), endpoint=True)):
            j *= speedup
            newx, newy = convert.lonlat2dymax(lon, lat)
            newx = int(newx*scale) - 1
            newy = int(newy*scale)
            try: dymaximg.putpixel((newx, newy), pix[i, j])
            # Sometimes a point won't map to an edge properly
            except IndexError: print('{{{:d}, {:d}}}'.format(newx, newy), end='')
    if verbose: print()
    dymaximg = ImageOps.flip(dymaximg) #it's upside down since putpixel flips too

    numpoints = im.size[0] * im.size[1] // speedup
    if verbose: print(':: mapped {:d} points to dymax projection @ {:.1f} pts/sec [{:.1f} secs total]'.format(numpoints, numpoints/(time.time()-start), time.time()-start))
    plt.figure(figsize=(20, 12), frameon=False)
    plt.gca().axis('off')
    if save:
        dymaximg.save(outFilename, format='PNG')
    if show:
        plt.tight_layout()
        plt.imshow(dymaximg)
        plt.show()
    else:
        plt.close()

def plot_face_hq(resolution:str='i', save:bool=False, show:bool=True, verbose:bool=True, dpi:int=300):
    '''Draw Dymax Triangles, All countries, and Meridians'''
    lonlat_islands, dymax_islands = io.get_coastlines(resolution)
    # select a face
    fdx = np.random.randint(0, 20)
    verts = constants.vert_indices[fdx]
    # contants.vertices
    # for vert in verts

    plt.figure(figsize=(20, 20))
    plt.title('Dymaxion Face {}'.format(fdx))
    # Plot Edges
    edges = convert.face2dymax(fdx)
    plt.plot(edges[:, 0], edges[:, 1], lw=1, alpha=.7, color='r')

    n = 1000 # meridian resolution
    ### Dymaxion Latitude Meridians
    lons = np.linspace(-180, 180, n)
    latgrid = np.linspace(-85, 85, 35)
    points = []
    # print(edges)
    for lat in latgrid:
        for lon in lons:
            point = convert.lonlat2dymax(lon, lat)
            if convert.raytrace(point[0], point[1], edges):
                points += [point]
    points = np.array(points)
    plt.plot(points[:, 0], points[:, 1], ',', color='k', alpha=.5)#,'.',lw=0)#,c=range(n))

    ### Dymaxion Longitude Meridians
    lats = np.linspace(-85, 85, n)
    longrid = np.linspace(-180, 175, 72)
    points = []
    start = time.time()
    for lon in longrid:
        for lat in lats:
            point = convert.lonlat2dymax(lon, lat)
            if convert.raytrace(point[0], point[1], edges):
                points += [point]
    points = np.array(points)
    plt.plot(points[:, 0], points[:, 1], ',', color='k', alpha=.5)#,'.',lw=0)#,c=range(n))

    ### Draw Landmasses
    patches = []
    for island in dymax_islands:
        island_closed = []
        for vertex in island:
            if convert.raytrace(vertex[0], vertex[1], edges):
                island_closed += [vertex]
        if len(island_closed) > 0:
            polygon = Polygon(np.array(island_closed), closed=True)#, closed=False, fill=False)
            patches += [polygon]
    p = PatchCollection(patches, alpha=1, linewidths=.8, edgecolor='k', facecolor='none')
    plt.gca().add_collection(p)
    if verbose: print(':: plotted', len(patches), 'coastlines')

    # Draw final figure bits
    plt.axis('off')
    plt.gca().set_aspect('equal')
    if save:
        plt.savefig('dymax_earthmeridianstriangles.png', bbox_inches='tight', dpi=dpi, transparent=True, pad_inches=0)
    if show:
        plt.tight_layout()
        plt.show()
    else:
        plt.close()

def run_examples(resolution:str='c', save:bool=False, show:bool=True, verbose:bool=True):
    '''
    Run all the examples in this file.
    The first part of this is really fast, the image conversion stuff
    '''
    if verbose: print('>> Running Dymax Projection Examples')
    plot_triangles(save=save, show=show)
    plot_triangles_meridians(resolution=resolution, save=save, show=show)
    plot_triangles_rectilinear(resolution=resolution, save=save, show=show)
    plot_lcd_triangles(resolution=resolution, save=save, show=show)
    plot_grid(save=save, show=show)
    plot_coastline_vectors(resolution=resolution, save=save, show=show)
    plot_face_hq(resolution=resolution, save=save, show=show)
    convert_rectimage_2_dymaximage(io.PKG_DATA+'bmng.jpg', 'dymax_bmng.png', save=save, show=show)
    convert_rectimage_2_dymaximage(io.PKG_DATA+'etopo1.jpg', 'dymax_etopo1.png', save=save, show=show)

if __name__ == '__main__':
    run_examples()
